'use strict';

// Reference: https://en.wikipedia.org/wiki/Lapse_rate

/**
 * Calculate the lapse rate between to height based on the initial location and the destination location
 * @param {Object.height + Object.temperature} from - initial location
 * @param {Object.height + Object.temperature} to - destination
 *
 * @return {Float} lapse rate - Unit are ()
 */
function envLapRate(from, to) {
  if (!from || !from.height || !from.temperature) throw new Error('Missing \'from\' input');
  if (!to || !to.height || !to.temperature) throw new Error('Missing \'to\' input');

  return -((to.height - from.height) / (to.temperature - from.temperature));
}

/**
 * Calculate moist adiabatic lapse rate based on air condition and water vapor condition
 */
function moistLapRate(waterValorPressure, airPressure, airTemperature) {
  const EARTH_GRAVITATIONAL_ACCELERATION = 9.8076; // m/s^2
  const WATER_VAPORIZATION_HEAT = 2501000; // J/KG
  const DRY_AIR_SPECIFIC_GAS_CONTENT = 287;  // J kg^−1 K^−1
  const WATER_VAPOR_SPECIFIC_GAS_CONTENT = 461.5;  // J kg^−1 K^−1
  const DIMENSIONLESS_RATIO = DRY_AIR_SPECIFIC_GAS_CONTENT / WATER_VAPOR_SPECIFIC_GAS_CONTENT;
  const CONSTANT_PRESSURE_DRY_AIR_HEAT = 1003.5; // J kg^1 K^1

  const mixingRatio = (DIMENSIONLESS_RATIO * waterValorPressure) / (airPressure - waterValorPressure);

  const lapseRate =
    EARTH_GRAVITATIONAL_ACCELERATION * (
      (1 + (WATER_VAPORIZATION_HEAT * mixingRatio) / (DRY_AIR_SPECIFIC_GAS_CONTENT * airTemperature))
      /
      (CONSTANT_PRESSURE_DRY_AIR_HEAT + (Math.pow(WATER_VAPORIZATION_HEAT, 2) * mixingRatio) / (WATER_VAPOR_SPECIFIC_GAS_CONTENT * Math.pow(airTemperature, 2)))
    );

  return lapseRate;
}

module.exports = {
  envLapRate,
  moistLapRate,
};
