'use strict';

const _haversineDistance = require('../haversine').haversineDistance;

/**
 * Decode a polyline string into an array of coordinates. Adapt from polyline.js
 * @see https://github.com/mapbox/polyline/blob/master/src/polyline.js
 *
 * @param {String} polyline - polyline String
 * @param {Int} precision - coordinates precision (number of decimal)
 *
 * @return {Array.Array.Number} coordinates
 *
 */
function decode(polyline, precision) {
  if (typeof polyline !== 'string') throw new Error(`Input polyline is not a string, got ${polyline}`);
  if (Number(precision) === precision && precision % 1 === 0) throw new Error(`Precision is not an Integer, got ${precision}`);

  let index = 0;
  let lat = 0;
  let lng = 0;
  const coordinates = [];
  let shift = 0;
  let result = 0;
  let byte = null;
  let latitude_change;
  let longitude_change;
  const factor = Math.pow(10, precision || 5);

  // Coordinates have variable length when encoded, so just keep
  // track of whether we've hit the end of the string. In each
  // loop iteration, a single coordinate is decoded.
  while (index < polyline.length) {

    // Reset shift, result, and byte
    byte = null;
    shift = 0;
    result = 0;

    do {
      byte = polyline.charCodeAt(index++) - 63;
      result |= (byte & 0x1f) << shift;
      shift += 5;
    } while (byte >= 0x20);

    latitude_change = ((result & 1) ? ~(result >> 1) : (result >> 1));

    shift = result = 0;

    do {
      byte = polyline.charCodeAt(index++) - 63;
      result |= (byte & 0x1f) << shift;
      shift += 5;
    } while (byte >= 0x20);

    longitude_change = ((result & 1) ? ~(result >> 1) : (result >> 1));

    lat += latitude_change;
    lng += longitude_change;

    coordinates.push([lat / factor, lng / factor]);
  }

  return coordinates;
}

/**
 * Python 2 rounding algorithm. Adapt from polyline.js
 * @see https://github.com/mapbox/polyline/blob/master/src/polyline.js
 *
 * @param {Float} value - Value to round
 * @return {Int} rounded - round value
 */
function _py2_round(value) {
    // Google's polyline algorithm uses the same rounding strategy as Python 2, which is different from JS for negative values
  return Math.floor(Math.abs(value) + 0.5) * Math.sign(value);
}

/**
 * Encode the input into a line of string. Adapt from polyline.js
 * @see https://github.com/mapbox/polyline/blob/master/src/polyline.js
 */
function _encode(current, previous, factor) {
  current = _py2_round(current * factor);
  previous = _py2_round(previous * factor);
  let coordinate = current - previous;
  coordinate <<= 1;

  if (current - previous < 0) {
    coordinate = ~coordinate;
  }

  let output = '';
  while (coordinate >= 0x20) {
    output += String.fromCharCode((0x20 | (coordinate & 0x1f)) + 63);
    coordinate >>= 5;
  }
  output += String.fromCharCode(coordinate + 63);
  return output;
}

/**
 * Encode the coordinates into a polyline. Adapt from polyline.js
 * @see https://github.com/mapbox/polyline/blob/master/src/polyline.js
 *
 * @param {Array.Array.Number} coordinates - an array of many coordinate arrays. Each coordinate array contain {Float} lat and a {Float} lon
 * @return {String} polyline - encoded polyline
 */
function encode(coordinates, precision) {
  if (!coordinates.length) { return ''; }

  const factor = Math.pow(10, precision || 5);
  let output = _encode(coordinates[0][0], 0, factor) + _encode(coordinates[0][1], 0, factor);

  for (let i = 1; i < coordinates.length; i++) {
    const a = coordinates[i];
    const b = coordinates[i - 1];
    output += _encode(a[0], b[0], factor);
    output += _encode(a[1], b[1], factor);
  }

  return output;
}

/**
 * Calculate the distance of the polyline. If radius is not provided, distance is flat, else distance is haversine distance
 * NOTE: Support flat surface and sphere
 *
 * @param {String} polyline - The polyline to calculate from
 * @param {Float} radius - The radius of the sphere, if not input, distance is calculated on flat surface
 * @param {Object.unit + } options - Options object. {String} unit.
 * @return {Float} length
 */
function length(polyline, radius, options) {
  if (typeof polyline !== 'string') throw new Error(`Input polyline is not a string, got ${polyline}`);
  if (radius && Number(radius) !== radius) throw new Error(`Input radius is not a number, got ${radius}`);

  const decodedPolyline = decode(polyline);
  let distance = 0;
  for (let i = 0; i < decodedPolyline.length - 1; i++) {

    const lat = [decodedPolyline[i][0], decodedPolyline[i][1]];
    const lon = [decodedPolyline[i + 1][0], decodedPolyline[i + 1][1]];

    if (radius) {
      // Haversine distance if there is radius
      distance += _haversineDistance(lat, lon, radius);
    } else {
      // Flat distance if there is not radius
      distance += Math.sqrt(Math.pow(lon[0] - lat[0], 2) - Math.pow(lon[1] - lat[1], 2));
    }
  }

  if (options && options.unit) {
    switch (options.unit) {
      case 'meter':
        return distance * 1000;
      case 'kilometer':
      default:
        return distance;
    }
  }

  return distance;
}

module.exports = {
  decode,
  encode,
  length,
};
